"""Creates and runs a temporary notebook to verify outputs"""

import json
from pathlib import Path
import re
import shutil
import subprocess
import sys
import tempfile
from . import utils


class NotebookRunFailedError(Exception):
    """
    raised when a notebook fails to run because the provided json is not a valid notebook
    """


class CellOutputMismatchError(Exception):
    """
    raised when the output of a code cell does not match the expected output
    """


class TempNotebook:
    """Creates and runs a temporary notebook to verify outputs"""

    def __init__(self, notebook_path: str):
        """
        Args:
            notebook_path (Union[str, pathlib.Path]): Path to the notebook file.
        """
        self.notebook_path = Path(notebook_path)
        with open(self.notebook_path, "r", encoding="UTF-8") as notebook_file:
            self.notebook_data = json.load(notebook_file)
        self.temp_dir = Path(tempfile.mkdtemp())
        self.output_notebook_path = self.temp_dir / "temp_notebook.ipynb"

    def run(self):
        """Runs the temporary notebook

        Raises:
            NotebookRunFailedError: if the notebook fails to run because
                the provided json is not a valid notebook
        """
        resp = subprocess.run(
            [
                sys.executable,
                "-m",
                "jupyter",
                "nbconvert",
                "--to",
                "notebook",
                "--execute",
                "--output",
                self.output_notebook_path,
                self.notebook_path,
            ],
            capture_output=True,
            check=False,
            encoding="utf-8",
            universal_newlines=True,
        )
        if resp.returncode != 0:
            stderr_without_ansi = re.sub(r"\x1B\[[0-?]*[ -/]*[@-~]", "", resp.stderr)
            raise NotebookRunFailedError(
                f"Notebook {self.output_notebook_path} failed to run.\n\n"
                f"Error message: {stderr_without_ansi}\n\n"
            )

        # get the json data from the saved notebook
        # file will be at filename.nbconvert.ipynb
        output_file_path = self.output_notebook_path
        output_data = utils.load_notebook_data(output_file_path)
        return output_data

    def compare_outputs(self, output_data: dict):
        """
        compares the outputs of cells of the temporary notebook to the cells
        of the output notebook

        Args:
            output_data (dict): the output notebook data

        Raises:
            CellOutputMismatchError: if the output of a cell does not match the expected output
        """
        code_cells = utils.get_code_cells(output_data)
        # if the first cell has the no-run comment, exit early
        if utils.cell_has_no_run_comment(code_cells[0]):
            return

        for cell_index, cell_data in enumerate(code_cells):
            # if the cell has a no-check-output comment, skip checking it
            if utils.cell_has_no_check_output_comment(cell_data):
                continue

            # check that the output cell matches the input cell
            if (
                cell_data["outputs"]
                != self.notebook_data["cells"][cell_index]["outputs"]
            ):
                raise CellOutputMismatchError(
                    f"Cell #{cell_index} output does not match the expected output.\n\n"
                    f"Cell contents: \n\n> {cell_data}"
                    "Expected output: \n\n> "
                    f"{self.notebook_data['cells'][cell_index]['outputs']}"
                )

    def check_notebook(self):
        """Runs the temporary notebook and compares the outputs"""
        output_data = self.run()
        self.compare_outputs(output_data)

    def __del__(self):
        """Deletes the temporary directory"""
        shutil.rmtree(self.temp_dir)

    def __enter__(self):
        """Returns the path to the temporary notebook"""
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Deletes the temporary directory"""
        del self



"""Forces notebooks to run cells sequentially and fails if cells were run out of order"""

import os
import pathlib
from . import temp_notebook
from . import utils


class NotebookCodeCellNotRunError(Exception):
    """Raised when a notebook code cell was not run"""


class NotebookRunOrderError(Exception):
    """Raised when a notebook is run out of order"""


class InvalidNotebookRunError(Exception):
    """Raised when any problems were identified with a notebook's run order"""


def notebook_is_in_virtualenv(notebook_path: pathlib.Path) -> bool:
    """
    Checks whether a notebook is in the current repo or is in the virtual environment's
    site-packages directory.

    Some dependencies may contain Jupyter notebooks in their site-packages directory for
    testing or documentation purposes. These notebooks should not be checked for run order.

    Args:
        notebook_path (pathlib.Path): Path to the notebook file.

    Returns:
        bool: True if the notebook is in the virtual environment's site-packages directory.
    """
    # if the notebook_path contains "site-packages", it is in the virtual environment
    # and should not be checked
    return "site-packages" in str(notebook_path)


def check_notebook_run_order(notebook_data: dict) -> None:
    """
    Checks that the notebook cells were run sequentially and fails if not.

    Args:
        notebook_data (dict): Notebook data in dictionary format.

    Raises:
        NotebookCodeCellNotRunError: If a code cell in the notebook was not run.
        NotebookRunOrderError: If the cells in the notebook were not run sequentially.
    """
    previous_cell_number = 0
    code_cells = utils.get_code_cells(notebook_data)
    for cell in code_cells:
        current_cell_number = cell["execution_count"]
        current_cell_source = cell["source"]
        # ignore empty cells
        if len(current_cell_source) > 0:
            if current_cell_number is None:
                raise NotebookCodeCellNotRunError(
                    f"Code cell was not run. The previous cell was #{previous_cell_number}. \n\n"
                    f"Cell contents: \n\n> {cell}"
                )
            if current_cell_number != previous_cell_number + 1:
                raise NotebookRunOrderError(
                    "Cells were not run sequentially. "
                    f"The cell that caused this error is #{current_cell_number} "
                    f"and the previous cell was #{previous_cell_number}. \n\n"
                    f"Cell contents: \n\n> {cell}"
                )
        previous_cell_number = current_cell_number


def check_single_notebook(notebook_path: str, no_run: bool = False):
    """Check a single notebook

    Args:
        notebook_path (str): Path to the notebook file.
        no_run (bool, optional): If True, do not run the notebook. Defaults to False.

    Raises:
        InvalidNotebookRunError: If any problems were identified with the notebook's run order.
    """
    notebook_path = pathlib.Path(notebook_path)
    notebook_data = utils.load_notebook_data(notebook_path)
    try:
        check_notebook_run_order(notebook_data)
        if not no_run:
            with temp_notebook.TempNotebook(notebook_path) as temp_nb:
                temp_nb.check_notebook()
    except (
        NotebookCodeCellNotRunError,
        NotebookRunOrderError,
        temp_notebook.NotebookRunFailedError,
        temp_notebook.CellOutputMismatchError,
    ) as error:
        raise InvalidNotebookRunError(
            f"Notebook {notebook_path} was not run in order.\n\n{error}\n\n"
        ) from error
    print(f"Notebook {notebook_path} was run correctly.")


def process_path(path: str, no_run: bool = False):
    """
    Processes a single path. Raises an exception if the path is invalid

    Args:
        path (str): Path to the notebook file or directory.
        no_run (bool, optional): If True, do not run the notebook. Defaults to False.

    Raises:
        ValueError: If the path is invalid.
    """

    if os.path.isdir(path):
        # Get all .ipynb files in the directory and its subdirectories
        for dirpath, _, filenames in os.walk(path):
            for filename in filenames:
                notebok_path = pathlib.Path(dirpath) / filename
                if filename.endswith(".ipynb") and not notebook_is_in_virtualenv(
                    notebok_path
                ):
                    check_single_notebook(
                        os.path.join(dirpath, filename),
                        no_run=no_run,
                    )
    elif path.endswith(".ipynb"):
        check_single_notebook(path, no_run=no_run)
    else:
        raise ValueError(
            f"Cannot check file {path}. "
            "Must be a path to a notebook file with the .ipynb extension, or a directory."
        )



"""Command-line interface for enforce_notebook_run_order"""

from typing import List
import warnings
import click
from .enforce_notebook_run_order import process_path


@click.command()
@click.argument("paths", nargs=-1, type=click.Path(exists=True), required=False)
@click.option(
    "--no-run",
    is_flag=True,
    help="Do not run the notebooks, only check the run order. "
    "This may miss some errors, but is useful for extremely long running notebooks. "
    "If you use this option, you should consider moving the long-running code to a "
    "separate task that runs separately from the notebook.",
)
def cli(paths: List[str] = None, no_run: bool = False):
    """
    Checks the run order of notebooks in the specified paths,
    or the entire repo if no paths are specified

    Args:
        paths: The paths to check. If no paths are specified, the current directory is checked.
        no_run: If True, the notebooks will not be run. This may miss some errors, but is useful
    """
    if no_run:
        warnings.warn(
            "The --no-run option will not catch all problems with notebooks. "
            "It is possible that the checks will pass, but the notebook was still run "
            "out of order. It is highly recommended to move any long-running code to a separate "
            "task that runs separately from the notebook."
        )

    if paths:
        for path in paths:
            process_path(path, no_run)
    else:
        # If no paths are provided, check the current directory
        process_path(".", no_run)



"""Contains shared functionality used across multiple modules"""

import json
from typing import Union


def load_notebook_data(notebook_path: str) -> dict:  # pragma: no cover
    """Loads the notebook data from the given path

    Args:
        notebook_path: Path to the notebook file.

    Returns:
        dict: Notebook data in dictionary format.
    """
    with open(notebook_path, "r", encoding="UTF-8") as notebook_file:
        notebook_data = json.load(notebook_file)
    return notebook_data


def get_code_cells(notebook_data: dict) -> list:  # pragma: no cover
    """Returns a list of code cells from the notebook data

    Args:
        notebook_data: Notebook data in dictionary format.

    Returns:
        list: List of code cells.
    """
    code_cells = []
    for cell in notebook_data["cells"]:
        if cell["cell_type"] == "code":
            code_cells.append(cell)
    return code_cells


def parse_cell_comment(cell: dict) -> Union[str, None]:
    """Returns the comment from the first line of the cell, if present

    Args:
        cell: The cell to parse

    Returns:
        str: The comment from the first line of the cell, if present
    """
    first_line = cell["source"][0]
    if first_line.startswith("#"):
        return first_line[1:].strip()
    return None


def cell_has_no_run_comment(cell_data: dict):
    """Returns True if the cell has a no-run comment

    Args:
        cell_data: The cell to check

    Returns:
        bool: True if the cell has a no-run comment
    """

    comment = parse_cell_comment(cell_data)
    if comment is None:
        return False
    return comment.lower() == "no-run"


def cell_has_no_check_output_comment(cell_data: dict):
    """Returns True if the cell has a no-check-output comment

    Args:
        cell_data: The cell to check

    Returns:
        bool: True if the cell has a no-check-output comment
    """
    comment = parse_cell_comment(cell_data)
    if comment is None:
        return False
    return comment.lower() == "no-check-output"
